from django.contrib import admin, messages
from django.contrib.admin.models import LogEntry
from django.contrib.admin.options import get_content_type_for_model
from django.contrib.admin.utils import unquote
from django.db import transaction
from django.shortcuts import redirect
from django.template.response import TemplateResponse
from django.utils.text import capfirst
from django.utils.translation import gettext_lazy as _

from actstream.models import any_stream
from simpel_admin.base import BaseModelAdmin


class AdminActivityMixin(BaseModelAdmin):
    def get_history_template(self):
        return self.object_history_template or [
            "admin/%s/%s/history.html" % (self.opts.app_label, self.opts.model_name),
            "admin/%s/%s_history.html" % (self.opts.app_label, self.opts.model_name),
            "admin/%s/object_history.html" % self.opts.app_label,
            "admin/object_history.html",
        ]

    def get_history_context(self, **kwargs):
        action_list = LogEntry.objects.filter(
            object_id=self.object.id,
            content_type=get_content_type_for_model(self.model)
        ).select_related().order_by('action_time')
        activity_list = any_stream(self.object).order_by("-timestamp")
        kwargs.update(
            {
                "opts": self.opts,
                "object": self.object,
                "title": self.get_inspect_title(self.object),
                "subtitle": None,
                "action_list": action_list,
                "activity_list": activity_list,
                "module_name": str(capfirst(self.opts.verbose_name_plural)),
                "preserved_filters": self.get_preserved_filters(self.request),
                "available_apps": admin.site.get_app_list(self.request),
                "has_change_permission": self.has_change_permission(self.request, self.object),
                "has_add_permission": self.has_add_permission(self.request),
                "has_delete_permission": self.has_delete_permission(self.request, self.object),
                "has_view_permission": self.has_view_permission(self.request, self.object),
                **self.admin_site.each_context(self.request),
            }
        )
        has_print_permission = getattr(self, "has_print_permission", None)
        if has_print_permission:
            kwargs["has_print_permission"] = has_print_permission(self.request, self.object)
        return kwargs

    def history_view(self, request, object_id, extra_context=dict()):
        """The 'history' admin view for this model. replaced by activity log"""
        # First check if the user can see this history.
        obj = self.get_object(request, unquote(object_id))
        if obj is None:
            return self._get_obj_does_not_exist_redirect(request, self.opts, object_id)

        self.request = request
        self.object = obj
        self.opts = self.model._meta

        if not self.has_view_or_change_permission(request, obj):
            messages.warning(request, _("You don't have permission to view %s history.") % obj)
            return redirect(self.get_changelist_url())

        # Then get the history for this object.
        context = self.get_history_context(**extra_context)
        request.current_app = self.admin_site.name
        return TemplateResponse(request, self.get_history_template(), context)

    def has_trash_permission(self, request, obj=None):
        return self.has_action_permission(request, "trash")

    def has_validate_permission(self, request, obj=None):
        return self.has_action_permission(request, "validate")

    def has_cancel_permission(self, request, obj=None):
        return self.has_action_permission(request, "cancel")

    def has_pay_permission(self, request, obj=None):
        return self.has_action_permission(request, "pay")

    def has_approve_permission(self, request, obj=None):
        return self.has_action_permission(request, "approve")

    def has_reject_permission(self, request, obj=None):
        return self.has_action_permission(request, "reject")

    def has_process_permission(self, request, obj=None):
        return self.has_action_permission(request, "process")

    def has_complete_permission(self, request, obj=None):
        return self.has_action_permission(request, "complete")

    def has_close_permission(self, request, obj=None):
        return self.has_action_permission(request, "close")

    def action_view(self, request, object_id, action=None, extra_context=None):
        # get objects
        if isinstance(object_id, self.model):
            obj = object_id
        else:
            obj = self.get_object(request, object_id)

        # Ignore action if status is not valid
        ignore_condition = getattr(obj, "%s_ignore_condition" % action)
        if ignore_condition:
            messages.warning(request, _("%s %s ignored.") % (action.title(), obj))
            return redirect(self.get_changelist_url())

        # Check action permission
        permission = getattr(self, "has_%s_permission" % action)
        if not permission(request, obj):
            messages.warning(request, _("You don't have %s permission.") % action)
            return redirect(self.get_changelist_url())

        # Perform action
        if request.method == "POST":
            try:
                with transaction.atomic():
                    object_action = getattr(obj, action)
                    object_action(request)
                messages.success(request, _("%s %s.") % (obj, action))
            except PermissionError as err:
                messages.error(request, err)
            return redirect(self.get_changelist_url())
        else:
            extra_context.update(
                {
                    "title": _("%s %s confirmation.") % (action.title(), self.opts.verbose_name, obj),
                    "object": obj,
                    "cancel_url": self.get_inspect_url(obj.id),
                }
            )
            return self.confirmation_view(request, extra_context=extra_context)

    @admin.action(
        permissions=["validate"],
        description=_("Validate selected %(verbose_name_plural)s"),
    )
    def validate_action(self, request, queryset):
        if queryset.count() != 1:
            err = _("Please select only one record.")
            messages.error(request, err)
            return
        obj = queryset.first()
        return self.action_view(request, obj, action="validate")

    @admin.action(
        permissions=["cancel"],
        description=_("Cancel selected %(verbose_name_plural)s"),
    )
    def cancel_action(self, request, queryset):
        if queryset.count() != 1:
            err = _("Please select only one record.")
            messages.error(request, err)
            return
        obj = queryset.first()
        return self.action_view(request, obj, action="cancel")

    @admin.action(
        permissions=["trash"],
        description=_("Trash selected %(verbose_name_plural)s"),
    )
    def trash_action(self, request, queryset):
        if queryset.count() != 1:
            err = _("Please select only one record.")
            messages.error(request, err)
            return
        obj = queryset.first()
        return self.action_view(request, obj, action="trash")

    @admin.action(
        permissions=["pay"],
        description=_("Pay selected %(verbose_name_plural)s"),
    )
    def pay_action(self, request, queryset):
        if queryset.count() != 1:
            err = _("Please select only one record.")
            messages.error(request, err)
            return
        obj = queryset.first()
        return self.action_view(request, obj, action="pay")

    @admin.action(
        permissions=["approve"],
        description=_("Approve selected %(verbose_name_plural)s"),
    )
    def approve_action(self, request, queryset):
        if queryset.count() != 1:
            err = _("Please select only one record.")
            messages.error(request, err)
            return
        obj = queryset.first()
        return self.action_view(request, obj, action="approve")

    @admin.action(
        permissions=["reject"],
        description=_("Reject selected %(verbose_name_plural)s"),
    )
    def reject_action(self, request, queryset):
        if queryset.count() != 1:
            err = _("Please select only one record.")
            messages.error(request, err)
            return
        obj = queryset.first()
        return self.action_view(request, obj, action="reject")

    @admin.action(
        permissions=["process"],
        description=_("Process selected %(verbose_name_plural)s"),
    )
    def process_action(self, request, queryset):
        if queryset.count() != 1:
            err = _("Please select only one record.")
            messages.error(request, err)
            return
        obj = queryset.first()
        return self.action_view(request, obj, action="process")

    @admin.action(
        permissions=["complete"],
        description=_("Complete selected %(verbose_name_plural)s"),
    )
    def complete_action(self, request, queryset):
        if queryset.count() != 1:
            err = _("Please select only one record.")
            messages.error(request, err)
            return
        obj = queryset.first()
        return self.action_view(request, obj, action="complete")

    @admin.action(
        permissions=["close"],
        description=_("Close selected %(verbose_name_plural)s"),
    )
    def close_action(self, request, queryset):
        if queryset.count() != 1:
            err = _("Please select only one record.")
            messages.error(request, err)
            return
        obj = queryset.first()
        return self.action_view(request, obj, action="close")



from django.db import models, transaction
from django.utils import timezone, translation

from actstream import action


from . import signals

_ = translation.gettext_lazy


class StatusMessage(models.Model):
    class Meta:
        abstract = True

    @property
    def opts(self):
        return self.__class__._meta

    def get_error_msg(self, action):
        msg = _("{}, {} is {}, it can't be {}.")
        return str(msg).format(
            self.opts.verbose_name,
            self,
            self.get_status_display(),
            action,
        )


class TrashMixin(StatusMessage, models.Model):

    date_trashed = models.DateTimeField(
        null=True,
        blank=True,
        editable=False,
        verbose_name=_("date trashed"),
    )

    class Meta:
        abstract = True

    @property
    def is_trash(self):
        """Check order status is trashed"""
        return self.status == self.TRASH

    @property
    def trash_ignore_condition(self):
        raise NotImplementedError

    @property
    def trash_valid_condition(self):
        raise NotImplementedError

    @transaction.atomic
    def trash(self, request=None):
        """Trash drafted order"""
        if self.trash_ignore_condition:
            return
        if self.trash_valid_condition:
            signals.pre_trash.send(sender=self.__class__, instance=self, actor=request.user, request=request)
            self.status = self.TRASH
            self.date_trashed = timezone.now()
            self.save()
            signals.post_trash.send(sender=self.__class__, instance=self, actor=request.user, request=request)
        else:
            raise PermissionError(self.get_error_msg("trash"))


class DraftMixin(StatusMessage, models.Model):

    date_drafted = models.DateTimeField(
        null=True,
        blank=True,
        editable=False,
        verbose_name=_("date drafted"),
    )

    class Meta:
        abstract = True

    @property
    def is_draft(self):
        """Check order status is draft"""
        return self.status == self.DRAFT

    @property
    def draft_ignore_condition(self):
        raise NotImplementedError

    @property
    def draft_valid_condition(self):
        raise NotImplementedError

    @transaction.atomic
    def draft(self, request=None):
        """Draft trashed"""
        if self.draft_ignore_condition:
            return
        if self.draft_valid_condition:
            signals.pre_draft.send(sender=self.__class__, instance=self, actor=request.user, request=request)
            self.status = self.DRAFT
            self.date_drafted = timezone.now()
            self.save()
            signals.post_draft.send(sender=self.__class__, instance=self, actor=request.user, request=request)
        else:
            raise PermissionError(self.get_error_msg("draft"))


class PendingMixin(StatusMessage, models.Model):

    date_pending = models.DateTimeField(
        null=True,
        blank=True,
        editable=False,
        verbose_name=_("date pending"),
    )

    class Meta:
        abstract = True

    @property
    def is_pending(self):
        """Check order status is pending"""
        return self.status == self.PENDING

    @transaction.atomic
    def pending(self, request=None):
        """pending trashed"""
        if self.is_pending:
            return
        if self.is_trash:
            self.status = self.PENDING
            self.date_pending = timezone.now()
            self.save()
        else:
            raise PermissionError(self.get_error_msg("pending"))


class ValidateMixin(StatusMessage, models.Model):

    date_validated = models.DateTimeField(
        null=True,
        blank=True,
        editable=False,
        verbose_name=_("date validated"),
    )

    class Meta:
        abstract = True

    @property
    def is_valid(self):
        """Check order status is valid"""
        return self.status == self.VALID

    @property
    def validate_ignore_condition(self):
        raise NotImplementedError

    @property
    def validate_valid_condition(self):
        raise NotImplementedError

    @transaction.atomic
    def validate(self, request=None):
        """Validate drafted order"""
        if self.validate_ignore_condition:
            return
        if self.validate_valid_condition:
            signals.pre_validate.send(sender=self.__class__, instance=self, actor=request.user, request=request)
            self.status = self.VALID
            self.date_validated = timezone.now()
            self.save()
            signals.post_validate.send(sender=self.__class__, instance=self, actor=request.user, request=request)
            action.send(request.user, verb="validate %s" % self._meta.verbose_name, action_object=self)
        else:
            raise PermissionError(self.get_error_msg("validated"))


class CancelMixin(StatusMessage, models.Model):

    date_canceled = models.DateTimeField(
        null=True,
        blank=True,
        editable=False,
        verbose_name=_("date canceled"),
    )

    class Meta:
        abstract = True

    @property
    def is_canceled(self):
        """Check order status is canceled"""
        return self.status == self.CANCELED

    @property
    def cancel_ignore_condition(self):
        raise NotImplementedError

    @property
    def cancel_valid_condition(self):
        raise NotImplementedError

    @transaction.atomic
    def cancel(self, request=None):
        """Validate drafted order"""
        if self.cancel_ignore_condition:
            return
        if self.cancel_valid_condition:
            signals.pre_cancel.send(sender=self.__class__, instance=self, actor=request.user, request=request)
            self.status = self.CANCELED
            self.date_canceled = timezone.now()
            self.save()
            signals.post_cancel.send(sender=self.__class__, instance=self, actor=request.user, request=request)
            action.send(request.user, verb="cancel %s" % self._meta.verbose_name, action_object=self)
        else:
            raise PermissionError(self.get_error_msg("canceled"))


class ApproveMixin(StatusMessage, models.Model):

    date_approved = models.DateTimeField(
        null=True,
        blank=True,
        editable=False,
        verbose_name=_("date approved"),
    )

    class Meta:
        abstract = True

    @property
    def is_approved(self):
        """Check order status is approved"""
        return self.status == self.APPROVED

    @property
    def approve_ignore_condition(self):
        raise NotImplementedError

    @property
    def approve_valid_condition(self):
        raise NotImplementedError

    @transaction.atomic
    def approve(self, request=None):
        """Approve valid order"""
        if self.approve_ignore_condition:
            return
        if self.approve_valid_condition:
            signals.pre_approve.send(sender=self.__class__, instance=self, actor=request.user, request=request)
            self.status = self.APPROVED
            self.date_approved = timezone.now()
            self.save()
            signals.post_approve.send(sender=self.__class__, instance=self, actor=request.user, request=request)
            action.send(request.user, verb="approve %s" % self._meta.verbose_name, action_object=self)
        else:
            raise PermissionError(self.get_error_msg("approved"))


class RejectMixin(StatusMessage, models.Model):

    date_rejected = models.DateTimeField(
        null=True,
        blank=True,
        editable=False,
        verbose_name=_("date rejected"),
    )

    class Meta:
        abstract = True

    @property
    def is_rejected(self):
        """Check order status is rejected"""
        return self.status == self.REJECTED

    @property
    def reject_ignore_condition(self):
        raise NotImplementedError

    @property
    def reject_valid_condition(self):
        raise NotImplementedError

    @transaction.atomic
    def reject(self, request=None):
        """Reject valid order"""
        if self.reject_ignore_condition:
            return
        if self.reject_valid_condition:
            signals.post_reject.send(sender=self.__class__, instance=self, actor=request.user, request=request)
            self.status = self.REJECTED
            self.date_rejected = timezone.now()
            self.save()
            signals.post_reject.send(sender=self.__class__, instance=self, actor=request.user, request=request)
            action.send(request.user, verb="reject %s" % self._meta.verbose_name, action_object=self)
        else:
            raise PermissionError(self.get_error_msg("rejected"))


class CompleteMixin(StatusMessage, models.Model):

    date_completed = models.DateTimeField(
        null=True,
        blank=True,
        editable=False,
        verbose_name=_("date completed"),
    )

    class Meta:
        abstract = True

    @property
    def is_complete(self):
        """Check order status is complete"""
        return self.status == self.COMPLETE

    @property
    def complete_ignore_condition(self):
        raise NotImplementedError

    @property
    def complete_valid_condition(self):
        raise NotImplementedError

    @transaction.atomic
    def complete(self, request=None):
        """Complete validated order"""
        if self.complete_ignore_condition:
            return
        if self.complete_valid_condition:
            signals.pre_complete.send(sender=self.__class__, instance=self, actor=request.user, request=request)
            self.status = self.COMPLETE
            self.date_completed = timezone.now()
            self.save()
            signals.post_complete.send(sender=self.__class__, instance=self, actor=request.user, request=request)
            action.send(request.user, verb="complete %s" % self._meta.verbose_name, action_object=self)
        else:
            raise PermissionError(self.get_error_msg("completed"))


class ProcessMixin(StatusMessage, models.Model):

    date_processed = models.DateTimeField(
        null=True,
        blank=True,
        editable=False,
        verbose_name=_("date processed"),
    )

    class Meta:
        abstract = True

    @property
    def is_processed(self):
        """Check order status is processed"""
        return self.status == self.PROCESSED

    @property
    def process_ignore_condition(self):
        raise NotImplementedError

    @property
    def process_valid_condition(self):
        raise NotImplementedError

    @transaction.atomic
    def process(self, request=None):
        """Process valid order"""
        if self.process_ignore_condition:
            return
        if self.process_valid_condition:
            signals.pre_process.send(sender=self.__class__, instance=self, actor=request.user, request=request)
            self.status = self.PROCESSED
            self.date_processed = timezone.now()
            self.save()
            signals.post_process.send(sender=self.__class__, instance=self, actor=request.user, request=request)
            action.send(request.user, verb="process %s" % self._meta.verbose_name, action_object=self)
        else:
            raise PermissionError(self.get_error_msg("processed"))


class PaidMixin(StatusMessage, models.Model):

    date_paid = models.DateTimeField(
        null=True,
        blank=True,
        editable=False,
        verbose_name=_("date paid"),
    )

    class Meta:
        abstract = True

    @property
    def is_paid(self):
        return self.status == self.PAID

    @property
    def pay_ignore_condition(self):
        raise NotImplementedError

    @property
    def pay_valid_condition(self):
        raise NotImplementedError

    @transaction.atomic
    def pay(self, request=None):
        """Paid pending order"""
        if self.pay_ignore_condition:
            return
        if self.pay_valid_condition:
            signals.pre_pay.send(sender=self.__class__, instance=self, actor=request.user, request=request)
            self.status = self.PAID
            self.date_paid = timezone.now()
            self.save()
            signals.post_pay.send(sender=self.__class__, instance=self, actor=request.user, request=request)
            action.send(request.user, verb="make payment fo %s" % self._meta.verbose_name, action_object=self)
        else:
            raise PermissionError(self.get_error_msg("paid"))


class CloseMixin(StatusMessage, models.Model):

    date_closed = models.DateTimeField(
        null=True,
        blank=True,
        editable=False,
        verbose_name=_("date closed"),
    )

    class Meta:
        abstract = True

    @property
    def is_closed(self):
        """Check object is closed"""
        return self.status == self.CLOSED

    @property
    def close_ignore_condition(self):
        raise NotImplementedError

    @property
    def close_valid_condition(self):
        raise NotImplementedError

    @transaction.atomic
    def close(self, request=None):
        """Close the order"""
        if self.close_ignore_condition:
            return
        if self.close_valid_condition:
            signals.pre_close.send(sender=self.__class__, instance=self, actor=request.user, request=request)
            self.status = self.CLOSED
            self.date_closed = timezone.now()
            self.save()
            signals.post_close.send(sender=self.__class__, instance=self, actor=request.user, request=request)
            action.send(request.user, verb="close %s" % self._meta.verbose_name, action_object=self)
        else:
            raise PermissionError(self.get_error_msg("closed"))


class ArchiveMixin(StatusMessage, models.Model):

    date_archived = models.DateTimeField(
        null=True,
        blank=True,
        editable=False,
        verbose_name=_("date archived"),
    )

    class Meta:
        abstract = True

    @property
    def is_archived(self):
        """Check order status is archived"""
        return self.status == self.ARCHIVED

    @property
    def archive_ignore_condition(self):
        raise NotImplementedError

    @property
    def archive_valid_condition(self):
        raise NotImplementedError

    @transaction.atomic
    def archive(self, request=None):
        """Archive the object"""
        if self.archive_ignore_condition:
            return
        if self.archive_valid_condition:
            signals.pre_archive.send(sender=self.__class__, instance=self, actor=request.user, request=request)
            self.status = self.ARCHIVED
            self.date_archived = timezone.now()
            self.save()
            signals.post_archive.send(sender=self.__class__, instance=self, actor=request.user, request=request)
            action.send(request.user, verb="archive %s" % self._meta.verbose_name, action_object=self)
        else:
            raise PermissionError(self.get_error_msg("archived"))


class StatusMixin(models.Model):
    """Base for status mixin used in sales order,
    warehouse transfer or invoice"""

    status = NotImplementedError

    class Meta:
        abstract = True

    @property
    def is_editable(self):
        """Check order is editable"""
        return self.is_trash or self.is_draft



from django.db.models.signals import ModelSignal

pre_trash = ModelSignal(use_caching=True)
post_trash = ModelSignal(use_caching=True)

pre_validate = ModelSignal(use_caching=True)
post_validate = ModelSignal(use_caching=True)

pre_cancel = ModelSignal(use_caching=True)
post_cancel = ModelSignal(use_caching=True)

pre_approve = ModelSignal(use_caching=True)
post_approve = ModelSignal(use_caching=True)

pre_confirm = ModelSignal(use_caching=True)
post_confirm = ModelSignal(use_caching=True)

pre_reject = ModelSignal(use_caching=True)
post_reject = ModelSignal(use_caching=True)

pre_complete = ModelSignal(use_caching=True)
post_complete = ModelSignal(use_caching=True)

pre_process = ModelSignal(use_caching=True)
post_process = ModelSignal(use_caching=True)

pre_invoicing = ModelSignal(use_caching=True)
post_invoicing = ModelSignal(use_caching=True)

pre_pay = ModelSignal(use_caching=True)
post_pay = ModelSignal(use_caching=True)

pre_close = ModelSignal(use_caching=True)
post_close = ModelSignal(use_caching=True)

pre_archive = ModelSignal(use_caching=True)
post_archive = ModelSignal(use_caching=True)



from django.db import models, transaction
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from . import signals


class ActionMixin(models.Model):
    status = NotImplementedError

    class Meta:
        abstract = True

    @property
    def is_editable(self):
        """Check order is editable"""
        return self.is_trash or self.is_draft

    @property
    def opts(self):
        return self.__class__._meta


class TrashAction(ActionMixin, models.Model):

    date_trashed = models.DateTimeField(
        null=True,
        blank=True,
        editable=False,
        verbose_name=_("date trashed"),
    )

    class Meta:
        abstract = True

    @property
    def is_trash(self):
        """Check order status is trashed"""
        return self.status == self.TRASH

    def trash(self, request=None):
        return self._trash(request)

    @transaction.atomic
    def _trash(self, request=None):
        self.allow_trash
        signals.pre_trash.send(
            sender=self.__class__,
            instance=self,
            actor=request.user,
            request=request,
        )
        self.status = self.TRASH
        self.date_trashed = timezone.now()
        self.save()
        signals.post_trash.send(
            sender=self.__class__,
            instance=self,
            actor=request.user,
            request=request,
        )


class DraftAction(ActionMixin, models.Model):

    date_drafted = models.DateTimeField(
        null=True,
        blank=True,
        editable=False,
        verbose_name=_("date drafted"),
    )

    class Meta:
        abstract = True

    @property
    def is_draft(self):
        """Check order status is draft"""
        return self.status == self.DRAFT

    def draft(self, request=None):
        return self._draft(request)

    @transaction.atomic
    def _draft(self, request=None):
        """Draft trashed"""
        signals.pre_draft.send(
            sender=self.__class__,
            instance=self,
            actor=request.user,
            request=request,
        )
        self.status = self.DRAFT
        self.date_drafted = timezone.now()
        self.save()
        signals.post_draft.send(
            sender=self.__class__,
            instance=self,
            actor=request.user,
            request=request,
        )


class PendingAction(ActionMixin, models.Model):

    date_pending = models.DateTimeField(
        null=True,
        blank=True,
        editable=False,
        verbose_name=_("date pending"),
    )

    class Meta:
        abstract = True

    @property
    def is_pending(self):
        """Check order status is pending"""
        return self.status == self.PENDING

    def pending(self, request=None):
        return self._pending(request)

    @transaction.atomic
    def _pending(self, request=None):
        """pending trashed"""
        if self.is_pending:
            return
        if self.is_trash:
            self.status = self.PENDING
            self.date_pending = timezone.now()
            self.save()
        else:
            raise PermissionError(self.get_error_msg("pending"))


class ValidateAction(ActionMixin, models.Model):

    date_validated = models.DateTimeField(
        null=True,
        blank=True,
        editable=False,
        verbose_name=_("date validated"),
    )

    class Meta:
        abstract = True

    @property
    def is_valid(self):
        """Check order status is valid"""
        return self.status == self.VALID

    def validate(self, request=None):
        return self._validate(request)

    @transaction.atomic
    def _validate(self, request=None):
        """Validate drafted order"""
        signals.pre_validate.send(
            sender=self.__class__,
            instance=self,
            actor=request.user,
            request=request,
        )
        self.status = self.VALID
        self.date_validated = timezone.now()
        self.save()
        signals.post_validate.send(
            sender=self.__class__,
            instance=self,
            actor=request.user,
            request=request,
        )


class CancelAction(ActionMixin, models.Model):

    date_canceled = models.DateTimeField(
        null=True,
        blank=True,
        editable=False,
        verbose_name=_("date canceled"),
    )

    class Meta:
        abstract = True

    @property
    def is_canceled(self):
        """Check order status is canceled"""
        return self.status == self.CANCELED

    def cancel(self, request=None):
        return self._cancel(request)

    @transaction.atomic
    def _cancel(self, request=None):
        signals.pre_cancel.send(
            sender=self.__class__,
            instance=self,
            actor=request.user,
            request=request,
        )
        self.status = self.CANCELED
        self.date_canceled = timezone.now()
        self.save()
        signals.post_cancel.send(
            sender=self.__class__,
            instance=self,
            actor=request.user,
            request=request,
        )


class ApproveAction(ActionMixin, models.Model):

    date_approved = models.DateTimeField(
        null=True,
        blank=True,
        editable=False,
        verbose_name=_("date approved"),
    )

    class Meta:
        abstract = True

    @property
    def is_approved(self):
        """Check order status is approved"""
        return self.status == self.APPROVED

    def approve(self, request=None):
        return self._approve(request)

    @transaction.atomic
    def _approve(self, request=None):
        signals.pre_approve.send(
            sender=self.__class__,
            instance=self,
            actor=request.user,
            request=request,
        )
        self.status = self.APPROVED
        self.date_approved = timezone.now()
        self.save()
        signals.post_approve.send(
            sender=self.__class__,
            instance=self,
            actor=request.user,
            request=request,
        )


class RejectAction(ActionMixin, models.Model):

    date_rejected = models.DateTimeField(
        null=True,
        blank=True,
        editable=False,
        verbose_name=_("date rejected"),
    )

    class Meta:
        abstract = True

    @property
    def is_rejected(self):
        """Check order status is rejected"""
        return self.status == self.REJECTED

    def reject(self, request=None):
        return self._reject(request)

    @transaction.atomic
    def _reject(self, request=None):
        """Reject valid order"""
        signals.post_reject.send(
            sender=self.__class__,
            instance=self,
            actor=request.user,
            request=request,
        )
        self.status = self.REJECTED
        self.date_rejected = timezone.now()
        self.save()
        signals.post_reject.send(
            sender=self.__class__,
            instance=self,
            actor=request.user,
            request=request,
        )


class CompleteAction(ActionMixin, models.Model):

    date_completed = models.DateTimeField(
        null=True,
        blank=True,
        editable=False,
        verbose_name=_("date completed"),
    )

    class Meta:
        abstract = True

    @property
    def is_complete(self):
        """Check order status is complete"""
        return self.status == self.COMPLETE

    def complete(self, request=None):
        return self._complete(request)

    @transaction.atomic
    def _complete(self, request=None):
        """Complete validated order"""
        signals.pre_complete.send(
            sender=self.__class__,
            instance=self,
            actor=request.user,
            request=request,
        )
        self.status = self.COMPLETE
        self.date_completed = timezone.now()
        self.save()
        signals.post_complete.send(
            sender=self.__class__,
            instance=self,
            actor=request.user,
            request=request,
        )


class ProcessAction(ActionMixin, models.Model):

    date_processed = models.DateTimeField(
        null=True,
        blank=True,
        editable=False,
        verbose_name=_("date processed"),
    )

    class Meta:
        abstract = True

    @property
    def is_processed(self):
        """Check order status is processed"""
        return self.status == self.PROCESSED

    def process(self, request=None):
        return self._process(request)

    @transaction.atomic
    def _process(self, request=None):
        """Process valid order"""
        signals.pre_process.send(
            sender=self.__class__,
            instance=self,
            actor=request.user,
            request=request,
        )
        self.status = self.PROCESSED
        self.date_processed = timezone.now()
        self.save()
        signals.post_process.send(
            sender=self.__class__,
            instance=self,
            actor=request.user,
            request=request,
        )


class PaidAction(ActionMixin, models.Model):

    date_paid = models.DateTimeField(
        null=True,
        blank=True,
        editable=False,
        verbose_name=_("date paid"),
    )

    class Meta:
        abstract = True

    @property
    def is_paid(self):
        return self.status == self.REJECTED

    def pay(self, request=None):
        return self._pay(request)

    @transaction.atomic
    def _pay(self, request=None):
        """Paid pending order"""
        signals.pre_pay.send(
            sender=self.__class__,
            instance=self,
            actor=request.user,
            request=request,
        )
        self.status = self.PAID
        self.date_paid = timezone.now()
        self.save()
        signals.post_pay.send(
            sender=self.__class__,
            instance=self,
            actor=request.user,
            request=request,
        )


class CloseAction(ActionMixin, models.Model):

    date_closed = models.DateTimeField(
        null=True,
        blank=True,
        editable=False,
        verbose_name=_("date closed"),
    )

    class Meta:
        abstract = True

    @property
    def is_closed(self):
        """Check object is closed"""
        return self.status == self.CLOSED

    def close(self, request=None):
        return self._close(request)

    @transaction.atomic
    def _close(self, request=None):
        """Close the order"""
        signals.pre_close.send(
            sender=self.__class__,
            instance=self,
            actor=request.user,
            request=request,
        )
        self.status = self.CLOSED
        self.date_closed = timezone.now()
        self.save()
        signals.post_close.send(
            sender=self.__class__,
            instance=self,
            actor=request.user,
            request=request,
        )


class ArchiveAction(ActionMixin, models.Model):

    date_archived = models.DateTimeField(
        null=True,
        blank=True,
        editable=False,
        verbose_name=_("date archived"),
    )

    class Meta:
        abstract = True

    @property
    def is_archived(self):
        """Check order status is archived"""
        return self.status == self.ARCHIVED

    def archive(self, request=None):
        return self._archive(request)

    @transaction.atomic
    def _archive(self, request=None):
        """Archive the object"""
        signals.pre_archive.send(
            sender=self.__class__,
            instance=self,
            actor=request.user,
            request=request,
        )
        self.status = self.ARCHIVED
        self.date_archived = timezone.now()
        self.save()
        signals.post_archive.send(
            sender=self.__class__,
            instance=self,
            actor=request.user,
            request=request,
        )



from django.apps import AppConfig
from django.conf import settings
from django.db.models.signals import post_migrate
from django.utils.translation import gettext_lazy as _


class SimpelActionsConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = "simpel_actions"
    label = "simpel_actions"
    verbose_name = _("Simpel Actions")

    def ready(self):
        post_migrate.connect(init_app, sender=self)


def init_app(sender, **kwargs):
    pass



from django.contrib import admin
from django.contrib.admin.templatetags.admin_urls import admin_urlname
from django.core.exceptions import ImproperlyConfigured
from django.urls import reverse

from .handlers import ActionHandler


class ActionableAdminMixin(admin.ModelAdmin):

    action_handler_class = ActionHandler

    def __init__(self, model, admin_site):
        super().__init__(model, admin_site)
        if self.action_handler_class is None:
            raise ImproperlyConfigured("%s missing action_handler_class!" % self.__class__.__name__)

    def get_action_redirect_url(self):
        return reverse(admin_urlname(self.opts, "changelist"))

    def perfom_action(self, action, request, queryset):
        try:
            for instance in queryset:
                self.action_handler.perform(action, instance, request)
            msg = "Successfully %s %s %s" % ("validate", queryset.count(), self.opts.verbose_name)
            self.message_user(request, msg, level="success")
        except Exception as err:
            self.message_user(request, err, level="error")



from django.contrib.auth import get_permission_codename
from django.utils.translation import gettext_lazy as _

from actstream import action



class ActionDoesNotExist(Exception):
    pass


class ActionHandler(object):

    action_logger_class = None

    def perform(self, action, instance, request=None, action_function=None):
        if action_function is None:
            try:
                action_function = getattr(self, "do_%s" % action)
            except Exception:
                raise ActionDoesNotExist(
                    _("%s doesn't have do_%s handler!") % (self.__class__.__name__, action),
                )
        return action_function(request, instance)

    def log(self, *args, **kwargs):
        """
        Log that an object action, override this action if using job queee.
        """
        action.send(**kwargs)

    def has_object_permission(self, request, code, obj=None):
        opts = obj.__class__._meta
        codename = get_permission_codename(code, opts)
        return request.user.has_perm("%s.%s" % (opts.app_label, codename))

    def has_action_permission(self, request, action, obj=None):
        perm_func = getattr(self, "has_%s_permission" % action, None)
        if perm_func is None:
            perm_func = self.has_object_permission
        return perm_func(request, action, obj)
